from in3py.model.in3_model import IN3Config
from in3py.model.in3_model import RPCRequest
from in3py.enums.in3 import EnumIn3Methods
from in3py.runner.in3_runner import In3Runner
from in3py.model.in3_model import ChainsDefinitions,Address
from typing import List


class IN3:

    def config(self, in3_config):
        rpc = RPCRequest()
        rpc.params = (in3_config,)
        rpc.method = EnumIn3Methods.CONFIG
        return In3Runner.call_in3_rpc(rpc)

    def checksum_address(self, address:Address, chain: bool):
        rpc = RPCRequest()
        rpc.method=EnumIn3Methods.CHECKSUM_ADDRESS
        rpc.params=(address,chain,)
        return In3Runner.call_in3_rpc(rpc)

    def abi_encode(self, method:str, args:list):
        rpc = RPCRequest()
        rpc.method=EnumIn3Methods.ABI_ENCODE
        rpc.params=(method,args,)
        return In3Runner.call_in3_rpc(rpc)

    def abi_decode(self, signature:str, data:str):
        rpc = RPCRequest()
        rpc.method=EnumIn3Methods.ABI_DECODE
        rpc.params=(signature,data,)
        return In3Runner.call_in3_rpc(rpc)








from in3py.model.in3_model import IN3Config

class In3Client:
    """
    In3Client instance for IN3 functions and Ethereum

    """

    def __init__(self, in3_config:IN3Config=None):
        from in3py.ethereum.ethereum import Ethereum
        from in3py.in3.in3 import IN3

        self.in3 = IN3()
        self.eth = Ethereum()
        if in3_config is not None:
            self.in3.config(in3_config=in3_config)




from enum import Enum, unique

@unique
class EnumsEthCall(Enum):

    WEB3_SHA3 = "web3_sha3"
    ACCOUNTS="eth_accounts"
    BALANCE="eth_getBalance"
    STORAGE_AT="eth_getStorageAt"
    GAS_PRICE="eth_gasPrice"
    BLOCK_NUMBER = "eth_blockNumber"
    BLOCK_BY_NUMBER="eth_getBlockByNumber"
    BLOCK_BY_HASH="eth_getBlockByHash"
    BLOCK_TRANSACTION_COUNT_BY_NUMBER = "eth_getBlockTransactionCountByNumber"
    BLOCK_TRANSACTION_COUNT_BY_HASH = "eth_getBlockTransactionCountByHash"
    TRANSACTION_BY_HASH="eth_getTransactionByHash"
    TRANSACTION_BY_BLOCKHASH_AND_INDEX="eth_getTransactionByBlockHashAndIndex"
    TRANSACTION_BY_BLOCKNUMBER_AND_INDEX="eth_getTransactionByBlockNumberAndIndex"
    TRANSACTION_RECEIPT="eth_getTransactionReceipt"
    PENDING_TRANSACTIONS="eth_pendingTransactions"
    TRANSACTION_COUNT = "eth_getTransactionCount"
    SEND_TRANSACTION = "eth_sendTransaction"
    SEND_RAW_TRANSACTION= "eth_sendRawTransaction"
    CALL = "eth_call"
    ESTIMATE_TRANSACTION="eth_estimateGas"
    UNCLE_COUNT_BY_BLOCK_HASH = "eth_getUncleCountByBlockHash"
    UNCLE_COUNT_BY_BLOCK_NUMBER = "eth_getUncleCountByBlockNumber"
    UNCLE_BY_BLOCKHASH_AND_INDEX="eth_getUncleByBlockHashAndIndex"
    UNCLE_BY_BLOCKNUMBER_AND_INDEX="eth_getUncleByBlockNumberAndIndex"
    CODE="eth_getCode"
    SIGN="eth_sign"
    NEW_FILTER="eth_newFilter"
    NEW_BLOCK_FILTER="eth_newBlockFilter"
    NEW_PENDING_TRANSACTION_FILTER="eth_newPendingTransactionFilter"
    UNINSTALL_FILTER="eth_uninstallFilter"
    FILTER_CHANGES="eth_getFilterChanges"
    FILTER_LOGS="eth_getFilterLogs"
    LOGS="eth_getLogs"

    def __str__(self):
        return self.value


@unique
class EnumIn3Methods(Enum):
    CHECKSUM_ADDRESS="in3_checksumAddress"
    SEND = "send"
    SIGN = "sign"
    CALL = "call"
    IN3_NODE_LIST = "in3_nodeList"
    IN3_SIGN = "in3_sign"
    IN3_STATS = "in3_stats"
    ABI_ENCODE = "in3_abiEncode"
    ABI_DECODE = "in3_abiDecode"
    PK_2_ADDRESS = "pk2address"
    PK_2_PUBLIC = "pk2public"
    ECRECOVER  = "ecrecover"
    CREATE_KEY = "createkey"
    KEY = "KEY"
    CONFIG = "in3_config"

    def __str__(self):
        return self.value


@unique
class EnumsBlockStatus(Enum):

    EARLIEST = "earliest"
    LATEST = "latest"
    PENDING = "pending"

    def __str__(self):
        return self.value





class IN3BaseException(Exception):
    """ In3 Base Exception """
    pass


class IN3NumberFormatException(IN3BaseException):
    """ It's not a valid number, please informe a valid format as 0x123 (Hex), 123(Hex without x) or an integer """
    pass


class PrivateKeyNotFoundException(IN3BaseException):
    """ Private Key not found, you can add it in  a local variable """
    pass


class AddressFormatException(IN3BaseException):
    """ The address is not correct, please check it again. """
    pass

class HashFormatException(IN3BaseException):

    """ The hash is not correct, please check the hash and try it again """
    pass


class In3RequestException(IN3BaseException):

    """ In3 Request expecetion  """
    pass




from in3py.model.in3_model import RPCRequest, RPCResponse
from in3py.enums.in3 import EnumsEthCall
from in3py.exceptions.in3_exceptions import In3RequestException
from in3py.runner.shared_library_runner import in3_create
from in3py.runner.shared_library_runner import in3_free
from in3py.runner.shared_library_runner import in3_exec
import json


class _In3Shared:

    def __init__(self):
        self.in3  = in3_create()

    def call(self, rpc):
        return in3_exec(self.in3, rpc)

    def release(self):
        in3_free(self.in3)


class In3Runner:

    __INSTANCE = None

    @staticmethod
    def get_instance():
        if In3Runner.__INSTANCE is None:
            In3Runner.__INSTANCE = _In3Shared()
        return In3Runner.__INSTANCE

    @staticmethod
    def call_in3_rpc(request: RPCRequest):
        response = In3Runner.get_instance().call(rpc=request.to_utf8())
        response_dict = json.loads(response.decode("utf-8").replace("\n"," "))

        rpc_response = RPCResponse()
        rpc_response.__dict__.update(response_dict)

        if rpc_response.error is not None:
            raise In3RequestException(rpc_response.error)

        # rpc_response.result = get_result_as_type(request.method, rpc_response.result)
        return get_result_as_type(request.method, rpc_response.result)


def get_result_as_type(method:EnumsEthCall, result:str):
    if method in [ EnumsEthCall.BALANCE, EnumsEthCall.GAS_PRICE, EnumsEthCall.BLOCK_NUMBER ]:
        return get_as_number(result)

    return result


def get_as_number(result:str):
    from in3py.model.in3_model import IN3Number
    return IN3Number(result)

# rpc=  RPCRequest()
# rpc.method = EnumsEthCall.BLOCK_NUMBER
#
# r = In3Runner.call_in3_rpc(rpc)
# print(r)


from in3py.runner.in3_runner import In3Runner
from in3py.model.in3_model import RPCRequest,Address,  IN3Number, Hash, Transaction, Filter
from in3py.enums.in3 import EnumsEthCall
from in3py.enums.in3 import EnumsBlockStatus


class Ethereum:
    """
        Ethereum module of In3 Client

    """
    def web3_sha3(self, data:str):
        """ Keccak-256 (not the standardized SHA3-256) of the given data.

            Args:
                data (str): the data to convert into a SHA3 hash.

            Returns:
                data (str): The SHA3 result of the given string.
        """

        rpc = RPCRequest(method=EnumsEthCall.WEB3_SHA3, params=(data,))
        return In3Runner.call_in3_rpc(rpc)

    def gas_price(self)->IN3Number:
        """ The current price per gas in wei.

            Args:
                -

            Returns:
                quantity (IN3Number): integer of the current gas price in wei.
        """

        rpc = RPCRequest(method=EnumsEthCall.GAS_PRICE)
        return In3Runner.call_in3_rpc(request=rpc)


    def block_number(self)->IN3Number:
        """ Returns the number of most recent block.

            Args:
                -
            Returns:
                block number (IN3Number) : number of most recent block

        """
        rpc = RPCRequest(method=EnumsEthCall.BLOCK_NUMBER)
        return In3Runner.call_in3_rpc(request=rpc)

    def get_balance(self, address:Address, number:[EnumsBlockStatus,IN3Number]=EnumsBlockStatus.LATEST)->IN3Number:
        """
        Returns the balance of the account of given address.

        :param Address address to check for balance. :
        :param number:
        :return:
            balance (IN3Number)
        """
        rpc = RPCRequest(method=EnumsEthCall.BALANCE, params=(address,number, ))
        return In3Runner.call_in3_rpc(request=rpc)

    def  get_storage_at(self, address:Address, position:IN3Number , number:[EnumsBlockStatus, IN3Number ]):
        rpc = RPCRequest(method=EnumsEthCall.STORAGE_AT, params=(address,position,number,))
        return In3Runner.call_in3_rpc(request=rpc)

    def get_transaction_count(self, address:Address, number:[IN3Number,EnumsBlockStatus]):
        rpc = RPCRequest(method=EnumsEthCall.TRANSACTION_COUNT, params=(address,number,))
        return In3Runner.call_in3_rpc(request=rpc)

    def get_block_transaction_count_by_hash(self, block_hash: Hash ):
        rpc = RPCRequest(method=EnumsEthCall.BLOCK_TRANSACTION_COUNT_BY_HASH, params=(block_hash,))
        return In3Runner.call_in3_rpc(request=rpc)

    def get_block_transaction_count_by_number(self, number: [IN3Number,EnumsBlockStatus] ):
        rpc = RPCRequest(method=EnumsEthCall.BLOCK_TRANSACTION_COUNT_BY_NUMBER, params=(number,))
        return In3Runner.call_in3_rpc(request=rpc)

    def get_uncle_count_by_block_hash(self, block_hash:Hash):
        rpc = RPCRequest(method=EnumsEthCall.UNCLE_COUNT_BY_BLOCK_HASH, params=(block_hash,))
        return In3Runner.call_in3_rpc(request=rpc)

    def get_uncle_count_by_block_number(self, number:[IN3Number,EnumsBlockStatus]):
        rpc = RPCRequest(method=EnumsEthCall.UNCLE_COUNT_BY_BLOCK_NUMBER, params=(number,))
        return In3Runner.call_in3_rpc(request=rpc)

    def get_code(self, address:Address, number:[IN3Number,  EnumsBlockStatus ]):
        rpc =RPCRequest(method=EnumsEthCall.CODE,  params=(address,number,) )
        return In3Runner.call_in3_rpc(request=rpc)

    def sign(self, address:Address, message:str):
        rpc = RPCRequest(method=EnumsEthCall.SIGN, params=(address, message,))
        return In3Runner.call_in3_rpc(request=rpc)

    def send_transaction(self, transaction:Transaction):
        rpc = RPCRequest(method=EnumsEthCall.SEND_TRANSACTION, params=(transaction,))
        return In3Runner.call_in3_rpc(request=rpc)

    def send_raw_transaction(self, data:str):
        rpc = RPCRequest(method=EnumsEthCall.SEND_RAW_TRANSACTION, params=(data,))
        return In3Runner.call_in3_rpc(request=rpc)

    def call(self, transaction:Transaction):
        rpc = RPCRequest(method=EnumsEthCall.CALL, params=(transaction,))
        return In3Runner.call_in3_rpc(rpc)

    def estimate_gas(self, transaction:Transaction):
        rpc = RPCRequest(method=EnumsEthCall.ESTIMATE_TRANSACTION, params=(transaction,))
        return In3Runner.call_in3_rpc(rpc)

    def get_block_by_hash(self, hash:Hash, is_full:bool=True):
        rpc = RPCRequest(method=EnumsEthCall.BLOCK_BY_HASH,params=(hash,is_full,))
        return In3Runner.call_in3_rpc(rpc)

    def get_block_by_number(self, number:[IN3Number, EnumsBlockStatus, str], is_full:bool=True):
        rpc = RPCRequest(method=EnumsEthCall.BLOCK_BY_NUMBER,params=(number,is_full,))
        return In3Runner.call_in3_rpc(rpc)

    def get_transaction_by_hash(self, hash:Hash):
        rpc = RPCRequest(method=EnumsEthCall.TRANSACTION_BY_HASH, params=(hash,))
        return In3Runner.call_in3_rpc(rpc)

    def get_transaction_by_block_hash_and_index(self,block_hash:Hash,index:IN3Number):
        rpc = RPCRequest(method=EnumsEthCall.TRANSACTION_BY_BLOCKHASH_AND_INDEX, params=(block_hash,index,))
        return  In3Runner.call_in3_rpc(rpc)

    def get_transaction_by_block_number_and_index(self, block_number:[IN3Number, EnumsBlockStatus], index:IN3Number):
        rpc = RPCRequest(method=EnumsEthCall.TRANSACTION_BY_BLOCKNUMBER_AND_INDEX,params=(block_number,index,))
        return In3Runner.call_in3_rpc(rpc)

    def get_transaction_receipt(self, hash: Hash):
        rpc = RPCRequest(method=EnumsEthCall.TRANSACTION_RECEIPT, params=(hash,))
        return In3Runner.call_in3_rpc(rpc)

    def pending_transactions(self):
        rpc = RPCRequest(method=EnumsEthCall.PENDING_TRANSACTIONS)
        return In3Runner.call_in3_rpc(rpc)

    def get_uncle_by_block_hash_and_index(self, block_hash:Hash, index:IN3Number):
        rpc = RPCRequest(method=EnumsEthCall.UNCLE_BY_BLOCKHASH_AND_INDEX, params=(block_hash,index,))
        return In3Runner.call_in3_rpc(rpc)

    def get_uncle_by_block_number_and_index(self, block_number:IN3Number, index:IN3Number):
        rpc = RPCRequest(method=EnumsEthCall.UNCLE_BY_BLOCKHASH_AND_INDEX, params=(block_number,index,))
        return In3Runner.call_in3_rpc(rpc)

    def new_filter(self, filter:Filter):
        rpc = RPCRequest(method=EnumsEthCall.NEW_FILTER, params=(filter, ))
        return In3Runner.call_in3_rpc(rpc)

    def new_block_filter(self):
        rpc=  RPCRequest(method=EnumsEthCall.NEW_BLOCK_FILTER)
        return In3Runner.call_in3_rpc(rpc)

    def new_pending_transaction_filter(self):
        rpc= RPCRequest(method=EnumsEthCall.NEW_PENDING_TRANSACTION_FILTER)
        return In3Runner.call_in3_rpc(rpc)

    def uninstall_filter(self, filter_id:IN3Number):
        rpc = RPCRequest(method=EnumsEthCall.UNINSTALL_FILTER, params=(filter_id,))
        return In3Runner.call_in3_rpc(rpc)

    def get_filter_changes(self, filter_id:IN3Number):
        rpc = RPCRequest(method=EnumsEthCall.FILTER_CHANGES,params=(filter_id,) )
        return In3Runner.call_in3_rpc(rpc)

    def get_filter_logs(self, filter_id:IN3Number):
        rpc =RPCRequest(method=EnumsEthCall.FILTER_LOGS, params=(filter_id,))
        return In3Runner.call_in3_rpc(rpc)

    def get_logs(self, filter:Filter):
        rpc =RPCRequest(method=EnumsEthCall.LOGS, params=(filter,))
        return In3Runner.call_in3_rpc(rpc)



























import json

from in3py.model.in3_model import RPCRequest
from in3py.model.in3_model import ChainsDefinitions
from in3py.model.in3_model import ModelUtils


def rpc_to_string(rpc:RPCRequest)->str:
    aux = rpc.__dict__
    aux["method"] = rpc.method.value
    aux["params"] = params_to_value(rpc.params)

    return json.dumps(aux)


def params_to_value(params):
    ret = []
    for p in params:
        ret.append(ModelUtils.to_json_rpc_param(p))
    return ret


def get_chain_by_id(id:str):
    for c in ChainsDefinitions.__members__:
        if ChainsDefinitions[c].value.chain_id == id:
            return ChainsDefinitions[c]





from distutils.core import setup
from setuptools import  find_packages

import os

def get_readme():
  from os import path
  this_directory = path.abspath(path.dirname(__file__))
  with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
      long_description = f.read()
  return long_description


envs = os.environ

version = envs.get("version", "0.0.15")
url = envs.get("url", "https://git.slock.it/in3/c/in3-python/")
download_url = envs.get("download_url","https://git.slock.it/in3/c/in3-python/-/archive/dev_0.1.0/in3-python-dev_0.1.0.tar.gz")
license = envs.get("license", "MIT")
description = envs.get("description", "In3 client for python")
keywords = envs.get("keywords", "in3,in3py").split(",")

name = envs.get("name", "in3")
author = envs.get("author", "Slockit")
author_email = envs.get("author_email", "developer@slock.it")
classifiers = envs.get("classifiers", "Development Status :: 3 - Alpha, Intended Audience :: Developers, Programming Language :: Python :: 3 ").split(",")


setup(
  name = name,
  # packages = find_packages(),
  packages = ["in3py"],
  include_package_data=True,
  version = version,
  license=license,
  description = description,
  author = author,
  author_email=author_email,
  url = url,
  # download_url=download_url,
  keywords = keywords,
  install_requires=[],
  classifiers=classifiers,
  long_description=get_readme(),
  long_description_content_type='text/markdown'
)

# 'Development Status :: 3 - Alpha',
# 'Intended Audience :: Developers',
# 'Topic :: Software Development :: Build Tools',
# 'License :: OSI Approved :: MIT License',
# 'Programming Language :: Python :: 3',
# 'Programming Language :: Python :: 3.4',
# 'Programming Language :: Python :: 3.5',
# 'Programming Language :: Python :: 3.6',


from ctypes import *
import requests

# the request struct
class Request(Structure):
    _fields_ = [("payload", POINTER(c_char)),
               ("urls", POINTER(POINTER(c_char))),
               ("urls_len", c_int),
               ("results", c_void_p)]

# the function called for each request
def transport(req):
    for i in range(0,req[0].urls_len):
        try:
            res = requests.post(string_at(req[0].urls[i]), data=string_at(req[0].payload), headers={'Content-type': 'application/json'})
            res.raise_for_status()
            lib.in3_req_add_response(req[0].results, i, False, res.content ,-1)
        except requests.exceptions.RequestException as e:
            lib.in3_req_add_response(req[0].results, i, True, e.strerror,-1)
    return 0


from pathlib import Path
from ctypes.util import find_library

path = Path(__file__).parent
lib_path = find_library(path.parent / "assets" / "libs" / "libin3") # macos "libin3.dylib"

lib = cdll.LoadLibrary(lib_path)
# free
lib._free_.argtypes = c_void_p,
lib._free_.restype = None
# new in3
lib.in3_new.argtypes = []
lib.in3_new.restype = c_void_p
#free in3
lib.in3_free.argtypes = c_void_p,
lib.in3_free.restype = None
# exec
lib.in3_client_exec_req.argtypes = [c_void_p,c_char_p]
lib.in3_client_exec_req.restype = c_void_p
# turn off logging
lib.in3_log_set_quiet_.argtypes = c_bool,
lib.in3_log_set_quiet_.restype = None
# declaring the function to report
lib.in3_req_add_response.argtypes = [c_void_p,c_int,c_bool,c_void_p,c_int]
lib.in3_req_add_response.restype = None
# define a transport function as callback
transport_fn = CFUNCTYPE(c_int, POINTER(Request))(transport)
# and set it as default
lib.in3_set_default_transport(transport_fn)
# register transport and verifiers (needed only once)
lib.in3_register_eth_full()
lib.in3_register_eth_api()
lib.in3_log_set_quiet_(True)


# create the in3-instance
def in3_create():
    return lib.in3_new()

# free the in3-instance
def in3_free(in3_ptr):
    lib.in3_free(in3_ptr)

# execute a rpc-request
def in3_exec(in3_ptr, rpc):
    ptr_res = lib.in3_client_exec_req(in3_ptr, rpc)
    result = cast(ptr_res, c_char_p).value
    lib._free_(ptr_res)
    return result

#
# # create a client
# in3 = in3_create()
#
# # run some rpc-requests
# print("current block:", in3_exec(in3, "{\"method\":\"eth_blockNumber\",\"params\":[]}".encode("utf-8") ))
#
# # clean up when done
# in3_free(in3)